# Copyright 2014 Dr. Greg M. Bernstein
""" A call blocking simulation example.

    Here we simulate an M/M/k/k queue and check the results against the Erlang
    loss formula.

    In this example we take a more OO approach and do not use global variables
    except for the running of the main simulation.
"""
import time
import random
import simpy
from erlang import erlangB


class RequestGenerator(object):
    """ Generates requests with exponential inter-arrival times.

        Parameters
        ----------
        env : simpy.Environment
            the simulation environment
        arrival_time : float
            the inter-arrival times of calls to the server resource.
        service_time : float
            the mean holding time of the calls.
        resource_pool : simpy.Resource
            the simpy.Resource representing the shared resource pool


    """
    def __init__(self, env, arrival_time, service_time, resource_pool):
        self.env = env
        self.resource_pool = resource_pool
        self.a_time = arrival_time
        self.s_time = service_time
        self.capacity = resource_pool.capacity
        self.request_count = 0
        self.requests_blocked = 0
        self.env.process(self.run())  # Add run() process to the simulation

    def run(self):
        while True:
            # Change the following line if you want a different inter-arrival
            # time distribution
            yield self.env.timeout(random.expovariate(1.0/self.a_time))
            self.request_count += 1
            if self.resource_pool.count == self.capacity:  # At capacity, call blocked!
                self.requests_blocked += 1
            else:  # create and start a new call process
                self.env.process(request_process(self.env, self.resource_pool, self.s_time))


def request_process(env, resource_pool, service_time):
    """ A model for a request.

        The request takes and holds a resource for an exponentially distributed
        service time.

        Parameters
        ----------
        env : simpy.Environment
            the simulation environment
        resource_pool : simpy.Resource
            the simpy.Resource the call will utilize
        service_time : float
            the mean service time
    """
    with resource_pool.request() as request:
        yield request
        yield env.timeout(random.expovariate(1.0/service_time))

if __name__ == '__main__':
    prog_start_time = time.time()
    arrival_rate = 17.0
    service_time = 1.3
    A = arrival_rate*service_time # The offered load in Erlangs
    k = 24  # The number of "server" or trunk resources
    env = simpy.Environment()  # The simulation environment.
    resource_pool = simpy.Resource(env, capacity=k)  # The "resource" pool
    call_gen = RequestGenerator(env, 1.0/arrival_rate, service_time, resource_pool)
    env.run(until=4000)
    print 'Total calls = {}, Dropped calls = {}, Blocking Probability = {}'.\
        format(call_gen.request_count, call_gen.requests_blocked, float(call_gen.requests_blocked)/call_gen.request_count)
    print 'Theoretical call blocking = {}'.format(erlangB(A, k))
    print 'Program running time in seconds: {}'.format(time.time()-prog_start_time)